import torch
import torch.nn.functional as F
INTERP_KWARGS = {'mode': 'trilinear', 'align_corners': True}


def smooth_kernel(kernel_size, sigma):
    meshgrids = torch.meshgrid([torch.arange(size, dtype=torch.float32, device=sigma.device) for size in kernel_size])
    kernel = 1
    for size, std, mgrid in zip(kernel_size, sigma, meshgrids):
        mean = (size - 1) / 2
        kernel *= 1 / (std * (2 * torch.pi)**.5) * torch.exp(-((mgrid - mean) / std) ** 2 / 2)
    return kernel / kernel.sum()


def jacobi_determinant(u, id_grid=None):
    gradient = jacobi_gradient(u, id_grid)
    dx, dy, dz = gradient[..., 2], gradient[..., 1], gradient[..., 0]
    jdet0 = dx[2] * (dy[1] * dz[0] - dy[0] * dz[1])
    jdet1 = dx[1] * (dy[2] * dz[0] - dy[0] * dz[2])
    jdet2 = dx[0] * (dy[2] * dz[1] - dy[1] * dz[2])
    jdet = jdet0 - jdet1 + jdet2
    return F.pad(jdet[None, None, 2:-2, 2:-2, 2:-2], (2, 2, 2, 2, 2, 2), mode='replicate')[0, 0]


def jacobi_gradient(u, id_grid=None):
    if id_grid is None:
        id_grid = create_grid(u.shape[1:4], u.device)
    x = 0.5 * (u + id_grid) * (torch.tensor(u.shape[1:4], device=u.device, dtype=u.dtype) - 1)
    window = torch.tensor([-.5, 0, .5], device=u.device)
    w = torch.zeros((3, 1, 3, 3, 3), device=u.device, dtype=u.dtype)
    w[2, 0, :, 1, 1] = window
    w[1, 0, 1, :, 1] = window
    w[0, 0, 1, 1, :] = window
    x = x.permute(4, 0, 1, 2, 3)
    x = F.conv3d(x, w)
    x = F.pad(x, (1, 1, 1, 1, 1, 1), mode='replicate')  # 'circular' for bfloat16
    return x.permute(0, 2, 3, 4, 1)


def create_grid(shape, device):
    return F.affine_grid(torch.eye(4, device=device)[None, :3], [1, 3, *shape], align_corners=INTERP_KWARGS['align_corners'])



import torch
import numpy as np
import torch.nn.functional as F
from tqdm import tqdm

from .metrics import LinearElasticity
from .utils import INTERP_KWARGS, create_grid, smooth_kernel
LIN_ELAST_FUNC = lambda x: LinearElasticity(mu=2., lam=1.)(x)


class SyNBase:
    def __init__(self, time_steps=7):
        self.time_steps = time_steps
        self._grid = None

    def apply_flows(self, x, y, v_xy, v_yx):
        half_flows = self.diffeomorphic_transform(torch.cat([v_xy, v_yx, -v_xy, -v_yx]))
        half_images = self.spatial_transform(torch.cat([x, y]), half_flows[:2])
        full_flows = self.composition_transform(half_flows[:2], half_flows[2:].flip(0))
        full_images = self.spatial_transform(torch.cat([x, y]), full_flows)
        images = {'xy_half': half_images[:1], 'yx_half': half_images[1:2],
                  'xy_full': full_images[:1], 'yx_full': full_images[1:2]}
        flows = {'xy_half': half_flows[:1], 'yx_half': half_flows[1:2],
                 'xy_full': full_flows[:1], 'yx_full': full_flows[1:2]}
        flows = {k: flow.permute(0, 2, 3, 4, 1) for k, flow in flows.items()}
        return images, flows

    def diffeomorphic_transform(self, v):
        v = v / (2 ** self.time_steps)
        for i in range(self.time_steps):
            v = v + self.spatial_transform(v, v)
        return v

    def composition_transform(self, v1, v2):
        return v2 + self.spatial_transform(v1, v2)

    def spatial_transform(self, x, v):
        if self._grid is None:
            self._grid = create_grid(v.shape[2:], x.device)
        return F.grid_sample(x, self._grid + v.permute(0, 2, 3, 4, 1), align_corners=True, padding_mode='reflection')


class SyNRegistration(SyNBase):
    def __init__(self, scales=(4, 2, 1), iterations=(30, 30, 10), learning_rate=1e-2, verbose=True,
                 dissimilarity_function=torch.nn.MSELoss(), regularization_function=LIN_ELAST_FUNC,
                 optimizer=torch.optim.Adam, sigma_img=.2, sigma_flow=.2, lambda_=2e-5, time_steps=7):
        super().__init__(time_steps=time_steps)
        self.scales = scales
        self.iterations = iterations
        self.learning_rates = [learning_rate] * len(scales) if isinstance(learning_rate, float) else learning_rate
        self.verbose = verbose
        self.dissimilarity_function = dissimilarity_function
        self.regularization_function = regularization_function
        self.optimizer = optimizer
        self.sigma_img = sigma_img
        self.sigma_flow = sigma_flow
        self.lambda_ = lambda_
        self.v_xy = None
        self.v_yx = None
        self._grid = None

    def __call__(self, moving, static, v_xy=None, v_yx=None, return_moved=True):
        if v_xy is None:
            v_xy = torch.zeros((moving.shape[0], 3, *moving.shape[2:]))
        if v_yx is None:
            v_yx = torch.zeros((static.shape[0], 3, *static.shape[2:]))
        self.v_xy = v_xy.type(static.dtype).to(static.device)
        self.v_yx = v_yx.type(static.dtype).to(static.device)
        for scale, iters, lr in zip(self.scales, self.iterations, self.learning_rates):
            moving_shape, static_shape = [s for s in moving.shape[2:]], [s for s in static.shape[2:]]
            shape = [int(round(s / scale)) for s in static_shape]
            self._grid = create_grid(shape, static.device)
            x = F.interpolate(moving, shape, **INTERP_KWARGS) if shape != moving_shape else moving.clone()
            y = F.interpolate(static, shape, **INTERP_KWARGS) if shape != static_shape else static.clone()
            if self.sigma_img:
                sigma_img = self.sigma_img * 200 / torch.tensor(shape).int()
                x = gauss_smoothing(x, sigma_img)
                y = gauss_smoothing(y, sigma_img)
            self.fit(x, y, iters, lr)
        self._grid = create_grid(static.shape[2:], static.device)
        if return_moved:
            images, flows = self.apply_flows(moving, static, self.v_xy, self.v_yx)
            return images['xy_full'], images['yx_full'], flows['xy_full'], flows['yx_full']

    def fit(self, x, y, iterations, learning_rate):
        v_xy = F.interpolate(self.v_xy, x.shape[2:], **INTERP_KWARGS)
        v_xy = torch.nn.Parameter(v_xy, requires_grad=True)
        v_yx = F.interpolate(self.v_yx, x.shape[2:], **INTERP_KWARGS)
        v_yx = torch.nn.Parameter(v_yx, requires_grad=True)
        sigma_flow = self.sigma_flow * torch.ones(3)
        optimizer = self.optimizer([v_xy, v_yx], learning_rate)
        progress_bar = tqdm(range(iterations), disable=not self.verbose)
        for _ in progress_bar:
            optimizer.zero_grad()
            images, flows = self.apply_flows(x, y, gauss_smoothing(v_xy, sigma_flow), gauss_smoothing(v_yx, sigma_flow))
            dissimilarity = (self.dissimilarity_function(x, images['yx_full']) +
                             self.dissimilarity_function(y, images['xy_full']) +
                             self.dissimilarity_function(images['yx_half'], images['xy_half']))
            regularization = (self.regularization_function(flows['yx_full']) +
                              self.regularization_function(flows['xy_full']))
            loss = dissimilarity + self.lambda_ * regularization
            progress_bar.set_description(f'Loss: {loss.item()}, '
                                         f'Dissimilarity: {dissimilarity.item()}, '
                                         f'Regularization: {regularization.item()}')
            loss.backward()
            optimizer.step()
        v_xy, v_yx = v_xy.detach(), v_yx.detach()
        v_xy, v_yx = gauss_smoothing(v_xy, sigma_flow), gauss_smoothing(v_yx, sigma_flow)
        self.v_xy = F.interpolate(v_xy, self.v_xy.shape[2:], **INTERP_KWARGS) if self.v_xy.shape != v_xy.shape else v_xy
        self.v_yx = F.interpolate(v_yx, self.v_yx.shape[2:], **INTERP_KWARGS) if self.v_yx.shape != v_yx.shape else v_yx


def gauss_smoothing(x, sigma):
    half_kernel_size = np.array(x.shape[2:]) // 50
    kernel_size = 1 + 2 * half_kernel_size.clip(min=1)
    kernel = smooth_kernel(kernel_size.tolist(), sigma).to(x.device)
    kernel = kernel[None, None].repeat(x.shape[1], 1, 1, 1, 1)
    x = F.pad(x, (kernel_size.repeat(2)[::-1] // 2).tolist(), mode='replicate')
    return F.conv3d(x.type(torch.float32), kernel, groups=x.shape[1]).type(x.dtype)



import torch
import torch.nn.functional as F
from tqdm import tqdm


class AffineRegistration:
    def __init__(self, scales=(4, 2), iterations=(500, 100), is_3d=True, learning_rate=1e-2,
                 verbose=True, dissimilarity_function=torch.nn.MSELoss(), optimizer=torch.optim.Adam,
                 init_translation=None, init_rotation=None, init_zoom=None, init_shear=None,
                 with_translation=True, with_rotation=True, with_zoom=True, with_shear=False,
                 align_corners=True, interp_mode=None, padding_mode='border'):
        self.scales = scales
        self.iterations = iterations[:len(scales)]
        self.is_3d = is_3d
        self.learning_rate = learning_rate
        self.verbose = verbose
        self.dissimilarity_function = dissimilarity_function
        self.optimizer = optimizer
        self.inits = (init_translation, init_rotation, init_zoom, init_shear)
        self.withs = (with_translation, with_rotation, with_zoom, with_shear)
        self.align_corners = align_corners
        self.interp_mode = 'trilinear' if is_3d else 'bilinear' if interp_mode is None else interp_mode
        self.padding_mode = padding_mode
        self._parameters = None
        self._loss = None

    def __call__(self, moving, static, return_moved=True):
        if len(moving.shape) - 4 != self.is_3d or len(static.shape) - 4 != self.is_3d:
            raise ValueError(f'Expected moving and static to be {4 + self.is_3d}D Tensors (2 + Spatial Dims.). '
                             f'Got size {moving.shape} and {static.shape}.')
        self._parameters = init_parameters(self.is_3d, len(static), static.device, *self.withs, *self.inits)
        interp_kwargs = {'mode': self.interp_mode, 'align_corners': self.align_corners}
        moving_ = F.interpolate(moving, static.shape[2:], **interp_kwargs)
        for scale, iters in zip(self.scales, self.iterations):
            moving_small = F.interpolate(moving_, scale_factor=1 / scale, **interp_kwargs)
            static_small = F.interpolate(static, scale_factor=1 / scale, **interp_kwargs)
            self._loss = self._fit(moving_small, static_small, iters)
        return self.transform(moving, static.shape[2:]).detach() if return_moved else None

    def _fit(self, moving, static, iterations):
        optimizer = self.optimizer(self._parameters, self.learning_rate)
        progress_bar = tqdm(range(iterations), disable=not self.verbose)
        for self.iter in progress_bar:
            optimizer.zero_grad()
            moved = self.transform(moving, static.shape[2:], with_grad=True)
            loss = self.dissimilarity_function(moved, static)
            progress_bar.set_description(f'Shape: {[*static.shape]}; Dissimiliarity: {loss.item()}')
            loss.backward()
            optimizer.step()
        return loss.item()

    def transform(self, moving, shape=None, with_grad=False):
        affine = self.get_affine(with_grad).type(moving.dtype)
        return affine_transform(moving, affine, shape, self.interp_mode, self.padding_mode, self.align_corners)

    def get_affine(self, with_grad=False):
        affine = compose_affine(*self._parameters)
        return affine if with_grad else affine.detach()


def affine_transform(x, affine, shape=None, mode='bilinear', padding_mode='border', align_corners=True):
    shape = x.shape[2:] if shape is None else shape
    grid = F.affine_grid(affine, [len(x), len(shape), *shape], align_corners)
    sample_mode = 'bilinear' if mode == 'trilinear' else mode  # grid_sample converts 'bi-' to 'trilinear' internally
    return F.grid_sample(x, grid, sample_mode, padding_mode, align_corners)


def init_parameters(is_3d=True, batch_size=1, device='cpu', with_translation=True, with_rotation=True, with_zoom=True,
                    with_shear=True, init_translation=None, init_rotation=None, init_zoom=None, init_shear=None):
    _check_parameter_shapes(init_translation, init_rotation, init_zoom, init_shear, is_3d, batch_size)
    n_dim = 2 + is_3d
    translation = torch.zeros(batch_size, n_dim).to(device) if init_translation is None else init_translation
    rotation = torch.stack(batch_size * [torch.eye(n_dim)]).to(device) if init_rotation is None else init_rotation
    zoom = torch.ones(batch_size, n_dim).to(device) if init_zoom is None else init_zoom
    shear = torch.zeros(batch_size, n_dim).to(device) if init_shear is None else init_shear
    params = [translation, rotation, zoom, shear]
    with_grad = [with_translation, with_rotation, with_zoom, with_shear]
    return [torch.nn.Parameter(param, requires_grad=grad) for param, grad in zip(params, with_grad)]


def compose_affine(translation, rotation, zoom, shear):
    _check_parameter_shapes(translation, rotation, zoom, shear, zoom.shape[-1] == 3, zoom.shape[0])
    square_matrix = torch.diag_embed(zoom)
    if zoom.shape[-1] == 3:
        square_matrix[..., 0, 1:] = shear[..., :2]
        square_matrix[..., 1, 2] = shear[..., 2]
    else:
        square_matrix[..., 0, 1] = shear[..., 0]
    square_matrix = rotation @ square_matrix
    return torch.cat([square_matrix, translation[:, :, None]], dim=-1)


def _check_parameter_shapes(translation, rotation, zoom, shear, is_3d=True, batch_size=1):
    n_dim = 2 + is_3d
    params = {'translation': translation, 'rotation': rotation, 'zoom': zoom, 'shear': shear}
    for name, param in params.items():
        if param is not None:
            desired_shape = (batch_size, n_dim, n_dim) if name == 'rotation' else (batch_size, n_dim)
            if param.shape != desired_shape:
                raise ValueError(f'Expected {name} to be size {desired_shape} since batch_size is {batch_size} '
                                 f'and is_3d is {is_3d} -> {2 + is_3d} dimensions. Got size {param.shape}.')



import torch
import torch.nn.functional as F

from .utils import create_grid, jacobi_gradient


def dice_loss(x1, x2):
    return 1 - dice_score(x1, x2)


def dice_score(x1, x2):
    dim = [2, 3, 4] if len(x2.shape) == 5 else [2, 3]
    inter = torch.sum(x1 * x2, dim=dim)
    union = torch.sum(x1 + x2, dim=dim)
    return (2. * inter / union).mean()


class LinearElasticity(torch.nn.Module):
    def __init__(self, mu=2., lam=1., refresh_id_grid=False):
        super(LinearElasticity, self).__init__()
        self.mu = mu
        self.lam = lam
        self.id_grid = None
        self.refresh_id_grid = refresh_id_grid

    def forward(self, u):
        if self.id_grid is None or self.refresh_id_grid:
            self.id_grid = create_grid(u.shape[1:4], u.device)
        gradients = jacobi_gradient(u, self.id_grid)
        u_xz, u_xy, u_xx = jacobi_gradient(gradients[None, 2], self.id_grid)
        u_yz, u_yy, u_yx = jacobi_gradient(gradients[None, 1], self.id_grid)
        u_zz, u_zy, u_zx = jacobi_gradient(gradients[None, 0], self.id_grid)
        e_xy = .5 * (u_xy + u_yx)
        e_xz = .5 * (u_xz + u_zx)
        e_yz = .5 * (u_yz + u_zy)
        sigma_xx = 2 * self.mu * u_xx + self.lam * (u_xx + u_yy + u_zz)
        sigma_xy = 2 * self.mu * e_xy
        sigma_xz = 2 * self.mu * e_xz
        sigma_yy = 2 * self.mu * u_yy + self.lam * (u_xx + u_yy + u_zz)
        sigma_yz = 2 * self.mu * e_yz
        sigma_zz = 2 * self.mu * u_zz + self.lam * (u_xx + u_yy + u_zz)
        return (sigma_xx ** 2 + sigma_xy ** 2 + sigma_xz ** 2 +
                sigma_yy ** 2 + sigma_yz ** 2 + sigma_zz ** 2).mean()


class NCC(torch.nn.Module):
    def __init__(self, kernel_size=7, epsilon_numerator=1e-5, epsilon_denominator=1e-5):
        super(NCC, self).__init__()
        self.kernel_size = kernel_size
        self.eps_nr = epsilon_numerator
        self.eps_dr = epsilon_denominator

    def forward(self, pred, targ):
        kernel = torch.ones([*targ.shape[:2]] + 3 * [self.kernel_size], device=targ.device)
        t_sum = F.conv3d(targ, kernel, padding=self.kernel_size // 2)
        p_sum = F.conv3d(pred, kernel, padding=self.kernel_size // 2)
        t2_sum = F.conv3d(targ ** 2, kernel, padding=self.kernel_size // 2)
        p2_sum = F.conv3d(pred ** 2, kernel, padding=self.kernel_size // 2)
        tp_sum = F.conv3d(targ * pred, kernel, padding=self.kernel_size // 2)
        cross = tp_sum - t_sum * p_sum / kernel.sum()
        t_var = F.relu(t2_sum - t_sum ** 2 / kernel.sum())
        p_var = F.relu(p2_sum - p_sum ** 2 / kernel.sum())
        cc = (cross ** 2 + self.eps_nr) / (t_var * p_var + self.eps_dr)
        return -torch.mean(cc)
